@@ -1,7 +1,7 @@
 /// @file htslib/hts.h
 /// Format-neutral I/O, indexing, and iterator API functions.
 /*
-    Copyright (C) 2012-2021 Genome Research Ltd.
+    Copyright (C) 2012-2022 Genome Research Ltd.
     Copyright (C) 2010, 2012 Broad Institute.
     Portions copyright (C) 2003-2006, 2008-2010 by Heng Li <lh3@live.co.uk>
 
@@ -366,6 +366,16 @@ enum hts_fmt_option {
     // name to the second field and insert a constructed <run>.<number>
     // name in its place.
     FASTQ_OPT_NAME2,
+
+    // Process the UMI tag.  Tag or Tag,tag,tag...
+    // On read, this converts the last read-name element (Illumina) to the tag.
+    // On write, it queries the tags in turn and copies the first found
+    // to the read name suffix, converting any non-alpha to "+".
+    FASTQ_OPT_UMI,
+
+    // Regex to use for matching read name.
+    // Def: "^[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:[^:]+:([^:#/]+)"
+    FASTQ_OPT_UMI_REGEX,
 };
 
 // Profile options for encoding; primarily used at present in CRAM
@@ -455,17 +465,21 @@ int hts_parse_opt_list(htsFormat *opt, const char *str);
 The input character may be either an IUPAC ambiguity code, '=' for 0, or
 '0'/'1'/'2'/'3' for a result of 1/2/4/8.  The result is encoded as 1/2/4/8
 for A/C/G/T or combinations of these bits for ambiguous bases.
+Additionally RNA U is treated as a T (8).
 */
+HTSLIB_EXPORT
 extern const unsigned char seq_nt16_table[256];
 
 /*! @abstract Table for converting a 4-bit encoded nucleotide to an IUPAC
 ambiguity code letter (or '=' when given 0).
 */
+HTSLIB_EXPORT
 extern const char seq_nt16_str[];
 
 /*! @abstract Table for converting a 4-bit encoded nucleotide to about 2 bits.
 Returns 0/1/2/3 for 1/2/4/8 (i.e., A/C/G/T), or 4 otherwise (0 or ambiguous).
 */
+HTSLIB_EXPORT
 extern const int seq_nt16_int[];
 
 /*!
@@ -486,7 +500,7 @@ const char *hts_version(void);
 // Immediately after release, bump ZZ to 90 to distinguish in-development
 // Git repository builds from the release; you may wish to increment this
 // further when significant features are merged.
-#define HTS_VERSION 101500
+#define HTS_VERSION 102300
 
 /*! @abstract Introspection on the features enabled in htslib
  *
@@ -673,7 +687,7 @@ int hts_set_opt(htsFile *fp, enum hts_fmt_option opt, ...);
   @param fp         The file handle
   @param delimiter  Unused, but must be '\n' (or KS_SEP_LINE)
   @param str        The line (not including the terminator) is written here
-  @return           Length of the string read;
+  @return           Length of the string read (capped at INT_MAX);
                     -1 on end-of-file; <= -2 on error
 */
 HTSLIB_EXPORT
@@ -1514,6 +1528,14 @@ static inline int hts_bin_level(int bin) {
     return l;
 }
 
+/**************************************
+ * Exposing the CRC32 implementation  *
+ * Either from zlib or libdeflate.    *
+ *************************************/
+HTSLIB_EXPORT
+uint32_t hts_crc32(uint32_t crc, const void *buf, size_t len);
+
+
 //! Compute the corresponding entry into the linear index of a given bin from
 //! a binning index
 /*!
@@ -1531,6 +1553,13 @@ static inline int hts_bin_bot(int bin, int n_lvls)
     return (bin - hts_bin_first(l)) << (n_lvls - l) * 3;
 }
 
+/// Compute the (0-based exclusive) maximum position covered by a binning index
+static inline hts_pos_t hts_bin_maxpos(int min_shift, int n_lvls)
+{
+    hts_pos_t one = 1;
+    return one << (min_shift + n_lvls * 3);
+}
+
 /**************
  * Endianness *
  **************/
