@@ -2,7 +2,7 @@
 /// High-level VCF/BCF variant calling file operations.
 /*
     Copyright (C) 2012, 2013 Broad Institute.
-    Copyright (C) 2012-2019 Genome Research Ltd.
+    Copyright (C) 2012-2020, 2022-2025 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -35,13 +35,15 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include <stdint.h>
 #include <limits.h>
-#include <assert.h>
 #include <errno.h>
 #include "hts.h"
 #include "kstring.h"
 #include "hts_defs.h"
 #include "hts_endian.h"
 
+/* Included only for backwards compatibility with e.g. bcftools 1.10 */
+#include <assert.h>
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -63,15 +65,25 @@ extern "C" {
 #define BCF_HT_STR  3
 #define BCF_HT_LONG (BCF_HT_INT | 0x100) // BCF_HT_INT, but for int64_t values; VCF only!
 
-#define BCF_VL_FIXED 0 // variable length
-#define BCF_VL_VAR   1
-#define BCF_VL_A     2
-#define BCF_VL_G     3
-#define BCF_VL_R     4
+// Values for INFO / FORMAT "Number=" fields
+#define BCF_VL_FIXED 0 ///< Integer defining a fixed number of items
+#define BCF_VL_VAR   1 ///< Generic variable length ("Number=.")
+#define BCF_VL_A     2 ///< One value per alternate allele
+#define BCF_VL_G     3 ///< One value for each possible genotype
+#define BCF_VL_R     4 ///< One value for each allele, including the reference
+
+// The following was introduced in VCFv4.4 and is only valid for FORMAT header lines
+#define BCF_VL_P     5 ///< One value for each allele value defined in GT
+
+// The following were introduced in VCFv4.5 and are only valid for FORMAT header lines
+#define BCF_VL_LA    6 ///< As BCF_VL_A, but only alt alleles listed in LAA are considered present
+#define BCF_VL_LG    7 ///< As BCF_VL_G, but only alt alleles listed in LAA are considered present
+#define BCF_VL_LR    8 ///< As BCF_VL_R, but only alt alleles listed in LAA are considered present
+#define BCF_VL_M     9 ///< One value for each possible base modification for the corresponding ChEBI ID
 
 /* === Dictionary ===
 
-   The header keeps three dictonaries. The first keeps IDs in the
+   The header keeps three dictionaries. The first keeps IDs in the
    "FILTER/INFO/FORMAT" lines, the second keeps the sequence names and lengths
    in the "contig" lines and the last keeps the sample names. bcf_hdr_t::dict[]
    is the actual hash table, which is opaque to the end users. In the hash
@@ -86,7 +98,7 @@ extern "C" {
 #define BCF_DT_SAMPLE   2
 
 // Complete textual representation of a header line
-typedef struct {
+typedef struct bcf_hrec_t {
     int type;       // One of the BCF_HL_* type
     char *key;      // The part before '=', i.e. FILTER/INFO/FORMAT/contig/fileformat etc.
     char *value;    // Set only for generic lines, NULL for FILTER/INFO, etc.
@@ -94,20 +106,20 @@ typedef struct {
     char **keys, **vals;    // The key=value pairs
 } bcf_hrec_t;
 
-typedef struct {
+typedef struct bcf_idinfo_t {
     uint64_t info[3];  // stores Number:20, var:4, Type:4, ColType:4 in info[0..2]
                        // for BCF_HL_FLT,INFO,FMT and contig length in info[0] for BCF_HL_CTG
     bcf_hrec_t *hrec[3];
     int id;
 } bcf_idinfo_t;
 
-typedef struct {
+typedef struct bcf_idpair_t {
     const char *key;
     const bcf_idinfo_t *val;
 } bcf_idpair_t;
 
 // Note that bcf_hdr_t structs must always be created via bcf_hdr_init()
-typedef struct {
+typedef struct bcf_hdr_t {
     int32_t n[3];           // n:the size of the dictionary block in use, (allocated size, m, is below to preserve ABI)
     bcf_idpair_t *id[3];
     void *dict[3];          // ID dictionary, contig dict and sample dict
@@ -121,6 +133,7 @@ typedef struct {
     int32_t m[3];          // m: allocated size of the dictionary block in use (see n above)
 } bcf_hdr_t;
 
+HTSLIB_EXPORT
 extern uint8_t bcf_type_shift[];
 
 /**************
@@ -135,19 +148,22 @@ extern uint8_t bcf_type_shift[];
 #define BCF_BT_FLOAT    5
 #define BCF_BT_CHAR     7
 
-#define VCF_REF      0
-#define VCF_SNP      1
-#define VCF_MNP      2
-#define VCF_INDEL    4
-#define VCF_OTHER    8
-#define VCF_BND     16    // breakend
-#define VCF_OVERLAP 32    // overlapping deletion, ALT=*
+#define VCF_REF         0
+#define VCF_SNP     (1<<0)
+#define VCF_MNP     (1<<1)
+#define VCF_INDEL   (1<<2)
+#define VCF_OTHER   (1<<3)
+#define VCF_BND     (1<<4)      // breakend
+#define VCF_OVERLAP (1<<5)      // overlapping deletion, ALT=*
+#define VCF_INS     (1<<6)      // implies VCF_INDEL
+#define VCF_DEL     (1<<7)      // implies VCF_INDEL
+#define VCF_ANY     (VCF_SNP|VCF_MNP|VCF_INDEL|VCF_OTHER|VCF_BND|VCF_OVERLAP|VCF_INS|VCF_DEL)       // any variant type (but not VCF_REF)
 
-typedef struct {
+typedef struct bcf_variant_t {
     int type, n;    // variant type and the number of bases affected, negative for deletions
-} variant_t;
+} bcf_variant_t;
 
-typedef struct {
+typedef struct bcf_fmt_t {
     int id;             // id: numeric tag id, the corresponding string is bcf_hdr_t::id[BCF_DT_ID][$id].key
     int n, size, type;  // n: number of values per-sample; size: number of bytes per-sample; type: one of BCF_BT_* types
     uint8_t *p;         // same as vptr and vptr_* in bcf_info_t below
@@ -155,7 +171,7 @@ typedef struct {
     uint32_t p_off:31, p_free:1;
 } bcf_fmt_t;
 
-typedef struct {
+typedef struct bcf_info_t {
     int key;        // key: numeric tag id, the corresponding string is bcf_hdr_t::id[BCF_DT_ID][$key].key
     int type;  // type: one of BCF_BT_* types
     union {
@@ -176,15 +192,15 @@ typedef struct {
 #define BCF1_DIRTY_FLT 4
 #define BCF1_DIRTY_INF 8
 
-typedef struct {
+typedef struct bcf_dec_t {
     int m_fmt, m_info, m_id, m_als, m_allele, m_flt; // allocated size (high-water mark); do not change
     int n_flt;  // Number of FILTER fields
     int *flt;   // FILTER keys in the dictionary
-    char *id, *als;     // ID and REF+ALT block (\0-seperated)
+    char *id, *als;     // ID and REF+ALT block (\0-separated)
     char **allele;      // allele[0] is the REF (allele[] pointers to the als block); all null terminated
     bcf_info_t *info;   // INFO
     bcf_fmt_t *fmt;     // FORMAT and individual sample
-    variant_t *var;     // $var and $var_type set only when set_variant_types called
+    bcf_variant_t *var; // $var and $var_type set only when set_variant_types called
     int n_var, var_type;
     int shared_dirty;   // if set, shared.s must be recreated on BCF output
     int indiv_dirty;    // if set, indiv.s must be recreated on BCF output
@@ -199,6 +215,22 @@ typedef struct {
 #define BCF_ERR_CTG_INVALID   32
 #define BCF_ERR_TAG_INVALID   64
 
+/// Get error description for bcf error code
+/** @param errorcode  The error code which is to be described
+    @param buffer     The buffer in which description to be added
+    @param maxbuffer  The size of buffer passed
+    @return NULL on invalid buffer; buffer on other cases
+
+The buffer will be an empty string when @p errorcode is 0.
+Description of errors present in code will be appended to @p buffer with ',' separation.
+The buffer has to be at least 4 characters long. NULL will be returned if it is smaller or when buffer is NULL.
+
+'...' will be appended if the description doesn't fit in the given buffer.
+ */
+
+HTSLIB_EXPORT
+const char *bcf_strerror(int errorcode, char *buffer, size_t maxbuffer);
+
 /*
     The bcf1_t structure corresponds to one VCF/BCF line. Reading from VCF file
     is slower because the string is first to be parsed, packed into BCF line
@@ -210,7 +242,7 @@ typedef struct {
     shared.s, indiv.s, etc.) are written directly by bcf_write, whereas a VCF
     line must be formatted in vcf_format.
  */
-typedef struct {
+typedef struct bcf1_t {
     hts_pos_t pos;  // POS
     hts_pos_t rlen; // length of REF
     int32_t rid;  // CHROM
@@ -302,6 +334,7 @@ typedef struct {
     typedef htsFile vcfFile;
     #define bcf_open(fn, mode) hts_open((fn), (mode))
     #define vcf_open(fn, mode) hts_open((fn), (mode))
+    #define bcf_flush(fp) hts_flush((fp))
     #define bcf_close(fp) hts_close(fp)
     #define vcf_close(fp) hts_close(fp)
 
@@ -359,6 +392,16 @@ typedef struct {
     HTSLIB_EXPORT
     int vcf_parse(kstring_t *s, const bcf_hdr_t *h, bcf1_t *v);
 
+    /**
+     * Complete the file opening mode, according to its extension.
+     * @param mode      Preallocated mode string to be completed.
+     * @param fn        File name to be opened.
+     * @param format    Format string (vcf|bcf|vcf.gz)
+     * @return          0 on success; -1 on failure
+     */
+    HTSLIB_EXPORT
+    int vcf_open_mode(char *mode, const char *fn, const char *format);
+
     /** The opposite of vcf_parse. It should rarely be called directly, see vcf_write */
     HTSLIB_EXPORT
     int vcf_format(const bcf_hdr_t *h, const bcf1_t *v, kstring_t *s);
@@ -388,7 +431,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     #define BCF_UN_INFO 4       // up to INFO
     #define BCF_UN_SHR  (BCF_UN_STR|BCF_UN_FLT|BCF_UN_INFO) // all shared information
     #define BCF_UN_FMT  8                           // unpack format and each sample
-    #define BCF_UN_IND  BCF_UN_FMT                  // a synonymo of BCF_UN_FMT
+    #define BCF_UN_IND  BCF_UN_FMT                  // a synonym of BCF_UN_FMT
     #define BCF_UN_ALL  (BCF_UN_SHR|BCF_UN_FMT)     // everything
     HTSLIB_EXPORT
     int bcf_unpack(bcf1_t *b, int which);
@@ -492,6 +535,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *
      *  The bcf_hdr_t struct returned by a successful call should be freed
      *  via bcf_hdr_destroy() when it is no longer needed.
+     *  @return NULL on failure, header otherwise
      */
     HTSLIB_EXPORT
     bcf_hdr_t *bcf_hdr_dup(const bcf_hdr_t *hdr);
@@ -509,6 +553,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *  bcf_hdr_merge() - copy header lines from src to dst, see also bcf_translate()
      *  @param dst: the destination header to be merged into, NULL on the first pass
      *  @param src: the source header
+     *  @return NULL on failure, header otherwise
      *
      *  Notes:
      *      - use as:
@@ -561,7 +606,8 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     int bcf_hdr_append(bcf_hdr_t *h, const char *line);
 
     HTSLIB_EXPORT
-    int bcf_hdr_printf(bcf_hdr_t *h, const char *format, ...);
+    int bcf_hdr_printf(bcf_hdr_t *h, const char *format, ...)
+    HTS_FORMAT(HTS_PRINTF_FMT, 2, 3);
 
     /** VCF version, e.g. VCFv4.2 */
     HTSLIB_EXPORT
@@ -589,6 +635,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *  @param n:        number of samples to keep
      *  @param samples:  names of the samples to keep
      *  @param imap:     mapping from index in @samples to the sample index in the original file
+     *  @return NULL on failure, header otherwise
      *
      *  Sample names not present in h0 are ignored. The number of unmatched samples can be checked
      *  by comparing n and bcf_hdr_nsamples(out_hdr).
@@ -600,7 +647,11 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     HTSLIB_EXPORT
     bcf_hdr_t *bcf_hdr_subset(const bcf_hdr_t *h0, int n, char *const* samples, int *imap);
 
-    /** Creates a list of sequence names. It is up to the caller to free the list (but not the sequence names) */
+    /**
+     *  Creates a list of sequence names. It is up to the caller to free the list (but not the sequence names).
+     *  NB: sequence name indexes returned by bcf_hdr_seqnames() may not correspond to bcf1_t.rid, use
+     *  bcf_hdr_id2name() or bcf_seqname() instead.
+     */
     HTSLIB_EXPORT
     const char **bcf_hdr_seqnames(const bcf_hdr_t *h, int *nseqs);
 
@@ -623,6 +674,23 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     HTSLIB_EXPORT
     int bcf_hdr_sync(bcf_hdr_t *h) HTS_RESULT_USED;
 
+    /**
+     * bcf_hdr_parse_line() - parse a single line of VCF textual header
+     * @param h     BCF header struct
+     * @param line  One or more lines of header text
+     * @param len   Filled out with length data parsed from 'line'.
+     * @return bcf_hrec_t* on success;
+     *         NULL on error or on end of header text.
+     *         NB: to distinguish error from end-of-header, check *len:
+     *           *len == 0 indicates @p line did not start with "##"
+     *           *len == -1 indicates failure, likely due to out of memory
+     *           *len > 0 indicates a malformed header line
+     *
+     * If *len > 0 on exit, it will contain the full length of the line
+     * including any trailing newline (this includes cases where NULL was
+     * returned due to a malformed line).  Callers can use this to skip to
+     * the next header line.
+     */
     HTSLIB_EXPORT
     bcf_hrec_t *bcf_hdr_parse_line(const bcf_hdr_t *h, const char *line, int *len);
     /// Convert a bcf header record to string form
@@ -634,6 +702,22 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     HTSLIB_EXPORT
     int bcf_hrec_format(const bcf_hrec_t *hrec, kstring_t *str);
 
+    /// Add a header record into a header
+    /**
+     *  @param hdr  Destination header
+     *  @param hrec Header record
+     *  @return 0 on success, -1 on failure
+     *
+     *  If this function returns success, ownership of @p hrec will have
+     *  been transferred to the header structure.  It may also have been
+     *  freed if it was a duplicate of a record already in the header.
+     *  Therefore the @p hrec pointer should not be used after a successful
+     *  return from this function.
+     *
+     *  If this function returns failure, ownership will not have been taken
+     *  and the caller is responsible for cleaning up @p hrec.
+     */
+
     HTSLIB_EXPORT
     int bcf_hdr_add_hrec(bcf_hdr_t *hdr, bcf_hrec_t *hrec);
 
@@ -716,15 +800,102 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     HTSLIB_EXPORT
     int bcf_translate(const bcf_hdr_t *dst_hdr, bcf_hdr_t *src_hdr, bcf1_t *src_line);
 
+    /// Get variant types in a BCF record
     /**
-     *  bcf_get_variant_type[s]()  - returns one of VCF_REF, VCF_SNP, etc
+     *  @param rec   BCF/VCF record
+     *  @return Types of variant present
+     *
+     *  The return value will be a bitwise-or of VCF_SNP, VCF_MNP,
+     *  VCF_INDEL, VCF_OTHER, VCF_BND or VCF_OVERLAP.  If will return
+     *  VCF_REF (i.e. 0) if none of the other types is present.
+     *  @deprecated Please use bcf_has_variant_types() instead
      */
     HTSLIB_EXPORT
     int bcf_get_variant_types(bcf1_t *rec);
 
+    /// Get variant type in a BCF record, for a given allele
+    /**
+     *  @param  rec        BCF/VCF record
+     *  @param  ith_allele Allele to check
+     *  @return Type of variant present
+     *
+     *  The return value will be one of VCF_REF, VCF_SNP, VCF_MNP,
+     *  VCF_INDEL, VCF_OTHER, VCF_BND or VCF_OVERLAP.
+     *  @deprecated Please use bcf_has_variant_type() instead
+     */
     HTSLIB_EXPORT
     int bcf_get_variant_type(bcf1_t *rec, int ith_allele);
 
+    /// Match mode for bcf_has_variant_types()
+    enum bcf_variant_match {
+        bcf_match_exact,   ///< Types present exactly match tested for
+        bcf_match_overlap, ///< At least one variant type in common
+        bcf_match_subset,  ///< Test set is a subset of types present
+    };
+
+    /// Check for presence of variant types in a BCF record
+    /**
+     *  @param rec      BCF/VCF record
+     *  @param bitmask  Set of variant types to test for
+     *  @param mode     Match mode
+     *  @return >0 if the variant types are present,
+     *           0 if not present,
+     *          -1 on error
+     *
+     *  @p bitmask should be the bitwise-or of the variant types (VCF_SNP,
+     *     VCF_MNP, etc.) to test for.
+     *
+     *  The return value is the bitwise-and of the set of types present
+     *  and @p bitmask.  Callers that want to check for the presence of more
+     *  than one type can avoid function call overhead by passing all the
+     *  types to be checked for in a single call to this function, in
+     *  bcf_match_overlap mode, and then check for them individually in the
+     *  returned value.
+     *
+     *  As VCF_REF is represented by 0 (i.e. the absence of other variants)
+     *  it should be tested for using
+     *    bcf_has_variant_types(rec, VCF_REF, bcf_match_exact)
+     *  which will return 1 if no other variant type is present, otherwise 0.
+     */
+    HTSLIB_EXPORT
+    int bcf_has_variant_types(bcf1_t *rec, uint32_t bitmask, enum bcf_variant_match mode);
+
+    /// Check for presence of variant types in a BCF record, for a given allele
+    /**
+     *  @param rec         BCF/VCF record
+     *  @param ith_allele  Allele to check
+     *  @param bitmask     Set of variant types to test for
+     *  @return >0 if one of the variant types is present,
+     *           0 if not present,
+     *          -1 on error
+     *
+     *  @p bitmask should be the bitwise-or of the variant types (VCF_SNP,
+     *     VCF_MNP, etc.) to test for, or VCF_REF on its own.
+     *
+     *  The return value is the bitwise-and of the set of types present
+     *  and @p bitmask.  Callers that want to check for the presence of more
+     *  than one type can avoid function call overhead by passing all the
+     *  types to be checked for in a single call to this function, and then
+     *  check for them individually in the returned value.
+     *
+     *  As a special case, if @p bitmask is VCF_REF (i.e. 0), the function
+     *  tests for an exact match.  The return value will be 1 if the
+     *  variant type calculated for the allele is VCF_REF, otherwise if
+     *  any other type is present it will be 0.
+     */
+    HTSLIB_EXPORT
+    int bcf_has_variant_type(bcf1_t *rec, int ith_allele, uint32_t bitmask);
+
+    /// Return the number of bases affected by a variant, for a given allele
+    /**
+     *  @param rec         BCF/VCF record
+     *  @param ith_allele  Allele index
+     *  @return The number of bases affected (negative for deletions),
+     *          or bcf_int32_missing on error.
+     */
+    HTSLIB_EXPORT
+    int bcf_variant_length(bcf1_t *rec, int ith_allele);
+
     HTSLIB_EXPORT
     int bcf_is_snp(bcf1_t *v);
 
@@ -756,7 +927,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     HTSLIB_EXPORT
     int bcf_has_filter(const bcf_hdr_t *hdr, bcf1_t *line, char *filter);
     /**
-     *  bcf_update_alleles() and bcf_update_alleles_str() - update REF and ALLT column
+     *  bcf_update_alleles() and bcf_update_alleles_str() - update REF and ALT column
      *  @alleles:           Array of alleles
      *  @nals:              Number of alleles
      *  @alleles_string:    Comma-separated alleles, starting with the REF allele
@@ -789,6 +960,13 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *  The @p string in bcf_update_info_flag() is optional,
      *  @p n indicates whether the flag is set or removed.
      *
+     *  Note that updating / removing END,SVLEN info tags will cause line->rlen
+     *  to be recalculated as a side-effect. If line->pos is being changed as
+     *  well, it is important that this is done before calling
+     *  bcf_update_info_int32() to update the END/SVLEN tag, otherwise rlen will be
+     *  set incorrectly.  If the new END value is less than or equal to
+     *  line->pos, a warning will be printed and line->rlen will be set to
+     *  the length of the REF allele for versions upto 4.3.
      */
     #define bcf_update_info_int32(hdr,line,key,values,n)   bcf_update_info((hdr),(line),(key),(values),(n),BCF_HT_INT)
     #define bcf_update_info_float(hdr,line,key,values,n)   bcf_update_info((hdr),(line),(key),(values),(n),BCF_HT_REAL)
@@ -833,6 +1011,8 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *  of fixed-length strings. In case of strings with variable length, shorter strings
      *  can be \0-padded. Note that the collapsed strings passed to bcf_update_format_char()
      *  are not \0-terminated.
+     *  With vcf4.5, rlen depends on format field LEN and rlen calculation uses v->pos as well.
+     *  So any change to pos be done before updating LEN that rlen calculated is correct.
      *
      *  Returns 0 on success or negative value on error.
      */
@@ -854,7 +1034,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     #define bcf_gt_unphased(idx)    (((idx)+1)<<1)
     #define bcf_gt_missing          0
     #define bcf_gt_is_missing(val)  ((val)>>1 ? 0 : 1)
-    #define bcf_gt_is_phased(idx)   ((idx)&1)
+    #define bcf_gt_is_phased(val)   ((val)&1)
     #define bcf_gt_allele(val)      (((val)>>1)-1)
 
     /** Conversion between alleles indexes to Number=G genotype index (assuming diploid, all 0-based) */
@@ -983,7 +1163,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *      int max_ploidy = ngt/nsmpl;
      *      for (i=0; i<nsmpl; i++)
      *      {
-     *        int32_t *ptr = gt + i*max_ploidy;
+     *        int32_t *ptr = gt_arr + i*max_ploidy;
      *        for (j=0; j<max_ploidy; j++)
      *        {
      *           // if true, the sample has smaller ploidy
@@ -1039,8 +1219,26 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
      *  bcf_hdr_id2name() - Translates numeric ID to sequence name
      */
     static inline int bcf_hdr_name2id(const bcf_hdr_t *hdr, const char *id) { return bcf_hdr_id2int(hdr, BCF_DT_CTG, id); }
-    static inline const char *bcf_hdr_id2name(const bcf_hdr_t *hdr, int rid) { return hdr->id[BCF_DT_CTG][rid].key; }
-    static inline const char *bcf_seqname(const bcf_hdr_t *hdr, bcf1_t *rec) { return hdr->id[BCF_DT_CTG][rec->rid].key; }
+    static inline const char *bcf_hdr_id2name(const bcf_hdr_t *hdr, int rid)
+    {
+        if ( !hdr || rid<0 || rid>=hdr->n[BCF_DT_CTG] ) return NULL;
+        return hdr->id[BCF_DT_CTG][rid].key;
+    }
+    static inline const char *bcf_seqname(const bcf_hdr_t *hdr, const bcf1_t *rec) {
+        return bcf_hdr_id2name(hdr, rec ? rec->rid : -1);
+    }
+
+    /** Return CONTIG name, or "(unknown)"
+
+        Like bcf_seqname(), but this function will never return NULL.  If
+        the contig name cannot be found (either because @p hdr was not
+        supplied or rec->rid was out of range) it returns the string
+        "(unknown)".
+    */
+    static inline const char *bcf_seqname_safe(const bcf_hdr_t *hdr, const bcf1_t *rec) {
+        const char *name = bcf_seqname(hdr, rec);
+        return name ? name : "(unknown)";
+    }
 
     /**
      *  bcf_hdr_id2*() - Macros for accessing bcf_idinfo_t
@@ -1060,7 +1258,7 @@ set to one of BCF_ERR* codes and must be checked before calling bcf_write().
     #define bcf_hdr_id2number(hdr,type,int_id)  ((hdr)->id[BCF_DT_ID][int_id].val->info[type]>>12)
     #define bcf_hdr_id2type(hdr,type,int_id)    (uint32_t)((hdr)->id[BCF_DT_ID][int_id].val->info[type]>>4 & 0xf)
     #define bcf_hdr_id2coltype(hdr,type,int_id) (uint32_t)((hdr)->id[BCF_DT_ID][int_id].val->info[type] & 0xf)
-    #define bcf_hdr_idinfo_exists(hdr,type,int_id)  ((int_id<0 || bcf_hdr_id2coltype(hdr,type,int_id)==0xf) ? 0 : 1)
+    #define bcf_hdr_idinfo_exists(hdr,type,int_id)  ((int_id)>=0 && (int_id)<(hdr)->n[BCF_DT_ID] && (hdr)->id[BCF_DT_ID][int_id].val && bcf_hdr_id2coltype((hdr),(type),(int_id))!=0xf)
     #define bcf_hdr_id2hrec(hdr,dict_type,col_type,int_id)    ((hdr)->id[(dict_type)==BCF_DT_CTG?BCF_DT_CTG:BCF_DT_ID][int_id].val->hrec[(dict_type)==BCF_DT_CTG?0:(col_type)])
     /// Convert BCF FORMAT data to string form
     /**
@@ -1270,7 +1468,14 @@ which works for both BCF and VCF.
 #define bcf_int16_missing    (-32767-1)      /* INT16_MIN */
 #define bcf_int32_missing    (-2147483647-1) /* INT32_MIN */
 #define bcf_int64_missing    (-9223372036854775807LL - 1LL)  /* INT64_MIN */
-#define bcf_str_missing      0x07
+
+// All of the above are values, which may occur multiple times in lists of
+// integers or lists of floating point.  Strings in VCF don't have
+// lists - a list of strings is just another (comma-separated) string.
+//
+// Hence bcf_str_missing is the whole string being missing rather than
+// an element of a list.  Ie a string of length zero: (0<<4)|BCF_BT_CHAR.
+#define bcf_str_missing      BCF_BT_CHAR
 
 // Limits on BCF values stored in given types.  Max values are the same
 // as for the underlying type.  Min values are slightly different as
@@ -1282,7 +1487,9 @@ which works for both BCF and VCF.
 #define BCF_MIN_BT_INT16 (-32760)      /* INT16_MIN + 8 */
 #define BCF_MIN_BT_INT32 (-2147483640) /* INT32_MIN + 8 */
 
+HTSLIB_EXPORT
 extern uint32_t bcf_float_vector_end;
+HTSLIB_EXPORT
 extern uint32_t bcf_float_missing;
 static inline void bcf_float_set(float *ptr, uint32_t value)
 {
@@ -1305,52 +1512,60 @@ static inline int bcf_float_is_vector_end(float f)
     return u.i==bcf_float_vector_end ? 1 : 0;
 }
 
+
+/**
+ *  bcf_format_gt_v2 - formats GT information on a string
+ *  @param hdr - bcf header, to get version
+ *  @param fmt - pointer to bcf format data
+ *  @param isample - position of interested sample in data
+ *  @param str - pointer to output string
+ *  Returns 0 on success and -1 on failure
+ *  This method is preferred over bcf_format_gt as this supports vcf4.4 and
+ *  prefixed phasing. Explicit / prefixed phasing for 1st allele is used only
+ *  when it is a must to correctly express phasing.
+ */
+HTSLIB_EXPORT
+int bcf_format_gt_v2(const bcf_hdr_t *hdr, bcf_fmt_t *fmt, int isample,
+    kstring_t *str) HTS_RESULT_USED;
+
 static inline int bcf_format_gt(bcf_fmt_t *fmt, int isample, kstring_t *str)
 {
-    uint32_t e = 0;
-    #define BRANCH(type_t, missing, vector_end) { \
-        type_t *ptr = (type_t*) (fmt->p + isample*fmt->size); \
-        int i; \
-        for (i=0; i<fmt->n && ptr[i]!=vector_end; i++) \
-        { \
-            if ( i ) e |= kputc("/|"[ptr[i]&1], str) < 0; \
-            if ( !(ptr[i]>>1) ) e |= kputc('.', str) < 0; \
-            else e |= kputw((ptr[i]>>1) - 1, str) < 0; \
-        } \
-        if (i == 0) e |= kputc('.', str) < 0; \
-    }
-    switch (fmt->type) {
-        case BCF_BT_INT8:  BRANCH(int8_t,  bcf_int8_missing, bcf_int8_vector_end); break;
-        case BCF_BT_INT16: BRANCH(int16_t, bcf_int16_missing, bcf_int16_vector_end); break;
-        case BCF_BT_INT32: BRANCH(int32_t, bcf_int32_missing, bcf_int32_vector_end); break;
-        case BCF_BT_NULL:  e |= kputc('.', str) < 0; break;
-        default: hts_log_error("Unexpected type %d", fmt->type); return -2;
-    }
-    #undef BRANCH
-    return e == 0 ? 0 : -1;
+    return bcf_format_gt_v2(NULL, fmt, isample, str);
 }
 
 static inline int bcf_enc_size(kstring_t *s, int size, int type)
 {
-    uint32_t e = 0;
-    if (size >= 15) {
-        e |= kputc(15<<4|type, s) < 0;
-        if (size >= 128) {
-            if (size >= 32768) {
-                int32_t x = size;
-                e |= kputc(1<<4|BCF_BT_INT32, s) < 0;
-                e |= kputsn((char*)&x, 4, s) < 0;
-            } else {
-                int16_t x = size;
-                e |= kputc(1<<4|BCF_BT_INT16, s) < 0;
-                e |= kputsn((char*)&x, 2, s) < 0;
-            }
+    // Most common case is first
+    if (size < 15) {
+        if (ks_resize(s, s->l + 1) < 0)
+            return -1;
+        uint8_t *p = (uint8_t *)s->s + s->l;
+        *p++ = (size<<4) | type;
+        s->l++;
+        return 0;
+    }
+
+    if (ks_resize(s, s->l + 6) < 0)
+        return -1;
+    uint8_t *p = (uint8_t *)s->s + s->l;
+    *p++ = 15<<4|type;
+
+    if (size < 128) {
+        *p++ = 1<<4|BCF_BT_INT8;
+        *p++ = size;
+        s->l += 3;
+    } else {
+        if (size < 32768) {
+            *p++ = 1<<4|BCF_BT_INT16;
+            i16_to_le(size, p);
+            s->l += 4;
         } else {
-            e |= kputc(1<<4|BCF_BT_INT8, s) < 0;
-            e |= kputc(size, s) < 0;
+            *p++ = 1<<4|BCF_BT_INT32;
+            i32_to_le(size, p);
+            s->l += 6;
         }
-    } else e |= kputc(size<<4|type, s) < 0;
-    return e == 0 ? 0 : -1;
+    }
+    return 0;
 }
 
 static inline int bcf_enc_inttype(long x)
@@ -1362,26 +1577,35 @@ static inline int bcf_enc_inttype(long x)
 
 static inline int bcf_enc_int1(kstring_t *s, int32_t x)
 {
-    uint32_t e = 0;
+    if (ks_resize(s, s->l + 5) < 0)
+        return -1;
+    uint8_t *p = (uint8_t *)s->s + s->l;
+
     if (x == bcf_int32_vector_end) {
-        e |= bcf_enc_size(s, 1, BCF_BT_INT8);
-        e |= kputc(bcf_int8_vector_end, s) < 0;
+        // An inline implementation of bcf_enc_size with size==1 and
+        // memory allocation already accounted for.
+        *p = (1<<4) | BCF_BT_INT8;
+        p[1] = bcf_int8_vector_end;
+        s->l+=2;
     } else if (x == bcf_int32_missing) {
-        e |= bcf_enc_size(s, 1, BCF_BT_INT8);
-        e |= kputc(bcf_int8_missing, s) < 0;
+        *p = (1<<4) | BCF_BT_INT8;
+        p[1] = bcf_int8_missing;
+        s->l+=2;
     } else if (x <= BCF_MAX_BT_INT8 && x >= BCF_MIN_BT_INT8) {
-        e |= bcf_enc_size(s, 1, BCF_BT_INT8);
-        e |= kputc(x, s) < 0;
+        *p = (1<<4) | BCF_BT_INT8;
+        p[1] = x;
+        s->l+=2;
     } else if (x <= BCF_MAX_BT_INT16 && x >= BCF_MIN_BT_INT16) {
-        int16_t z = x;
-        e |= bcf_enc_size(s, 1, BCF_BT_INT16);
-        e |= kputsn((char*)&z, 2, s) < 0;
+        *p = (1<<4) | BCF_BT_INT16;
+        i16_to_le(x, p+1);
+        s->l+=3;
     } else {
-        int32_t z = x;
-        e |= bcf_enc_size(s, 1, BCF_BT_INT32);
-        e |= kputsn((char*)&z, 4, s) < 0;
+        *p = (1<<4) | BCF_BT_INT32;
+        i32_to_le(x, p+1);
+        s->l+=5;
     }
-    return e == 0 ? 0 : -1;
+
+    return 0;
 }
 
 /// Return the value of a single typed integer.
@@ -1390,10 +1614,10 @@ static inline int bcf_enc_int1(kstring_t *s, int32_t x)
     @param[out] q    Location to store an updated value for p
     @return The integer value, or zero if @p type is not valid.
 
-If @p type is not one of BCF_BT_INT8, BCF_BT_INT16 or BCF_BT_INT32, zero
-will be returned and @p *q will not be updated.  Otherwise, the integer
-value will be returned and @p *q will be set to the memory location
-immediately following the integer value.
+If @p type is not one of BCF_BT_INT8, BCF_BT_INT16, BCF_BT_INT32 or
+BCF_BT_INT64, zero will be returned and @p *q will not be updated.
+Otherwise, the integer value will be returned and @p *q will be set
+to the memory location immediately following the integer value.
 
 Cautious callers can detect invalid type codes by checking that *q has
 actually been updated.
@@ -1411,7 +1635,7 @@ static inline int64_t bcf_dec_int1(const uint8_t *p, int type, uint8_t **q)
         *q = (uint8_t*)p + 4;
         return le_to_i32(p);
     } else if (type == BCF_BT_INT64) {
-        *q = (uint8_t*)p + 4;
+        *q = (uint8_t*)p + 8;
         return le_to_i64(p);
     } else { // Invalid type.
         return 0;
